<?php

declare(strict_types=1);

namespace App\Context\Content\Application\Command\Question\UpdateQuestion;

use App\Context\Content\Domain\Model\Level\LevelId;
use App\Context\Content\Domain\Model\Question\Exception\DescriptionEqualsCurrentException;
use App\Context\Content\Domain\Model\Question\Exception\SlugEqualsCurrentException;
use App\Context\Content\Domain\Model\Question\Exception\TitleEqualsCurrentException;
use App\Context\Content\Domain\Model\Question\Policy\QuestionEditPolicy;
use App\Context\Content\Domain\Model\Question\QuestionDescription;
use App\Context\Content\Domain\Model\Question\QuestionId;
use App\Context\Content\Domain\Model\Question\QuestionRepositoryPersistence;
use App\Context\Content\Domain\Model\Question\QuestionRepositoryReader;
use App\Context\Content\Domain\Model\Question\QuestionSlug;
use App\Context\Content\Domain\Model\Question\QuestionTitle;
use App\Context\Content\Domain\Model\Question\Specification\QuestionSlugUniqueSpec;
use App\Context\Content\Domain\Model\Skill\SkillId;
use App\Context\Shared\Domain\Bus\Command\CommandHandler;

class UpdateQuestionHandler implements CommandHandler
{
    public function __construct(
        private readonly QuestionRepositoryReader $questionReader,
        private readonly QuestionRepositoryPersistence $persistence,
        private readonly QuestionSlugUniqueSpec $questionSlugUniqueSpec,
        private readonly QuestionEditPolicy $questionEditPolicy,
    ) {
    }

    public function __invoke(UpdateQuestionCommand $command): UpdateQuestionCommandResponse
    {
        $skillId = new SkillId($command->skillId);
        $levelId = new LevelId($command->levelId);
        $questionId = new QuestionId($command->questionId);
        $question = $this->questionReader->getActiveById($skillId, $levelId, $questionId);

        if (null !== $command->title) {
            try {
                $question->changeTitle($this->questionEditPolicy, new QuestionTitle($command->title));
            } catch (TitleEqualsCurrentException) {
            }
        }

        if (null !== $command->slug) {
            try {
                $question->changeSlug(
                    $this->questionEditPolicy,
                    $this->questionSlugUniqueSpec,
                    new QuestionSlug($command->slug)
                );
            } catch (SlugEqualsCurrentException) {
            }
        }

        if (null !== $command->description) {
            try {
                $question->changeDescription($this->questionEditPolicy, new QuestionDescription($command->description));
            } catch (DescriptionEqualsCurrentException) {
            }
        }

        $this->persistence->persist($question);

        return new UpdateQuestionCommandResponse($question->getId());
    }
}
