<?php

declare(strict_types=1);

namespace App\Context\Content\Domain\Model\Level;

use App\Context\Content\Domain\Model\Level\Event\LevelCreated;
use App\Context\Content\Domain\Model\Level\Event\LevelDescriptionChanged;
use App\Context\Content\Domain\Model\Level\Event\LevelModeratorAssigned;
use App\Context\Content\Domain\Model\Level\Event\LevelModeratorRemoved;
use App\Context\Content\Domain\Model\Level\Event\LevelOwnerChanged;
use App\Context\Content\Domain\Model\Level\Event\LevelRemoved;
use App\Context\Content\Domain\Model\Level\Event\LevelRestored;
use App\Context\Content\Domain\Model\Level\Event\LevelSlugChanged;
use App\Context\Content\Domain\Model\Level\Event\LevelTitleChanged;
use App\Context\Content\Domain\Model\Level\Exception\DescriptionEqualsCurrentException;
use App\Context\Content\Domain\Model\Level\Exception\LevelAlreadyRemoveException;
use App\Context\Content\Domain\Model\Level\Exception\LevelModeratorAlreadyAssignedException;
use App\Context\Content\Domain\Model\Level\Exception\LevelModeratorNotAssignedException;
use App\Context\Content\Domain\Model\Level\Exception\LevelNotRemovedException;
use App\Context\Content\Domain\Model\Level\Exception\LevelSlugAlreadyExist;
use App\Context\Content\Domain\Model\Level\Exception\OwnerEqualsCurrentException;
use App\Context\Content\Domain\Model\Level\Exception\Policy\LevelCreatePolicyException;
use App\Context\Content\Domain\Model\Level\Exception\Policy\LevelEditPolicyException;
use App\Context\Content\Domain\Model\Level\Exception\Policy\LevelRemovePolicyException;
use App\Context\Content\Domain\Model\Level\Exception\Policy\LevelRestorePolicyException;
use App\Context\Content\Domain\Model\Level\Exception\SlugEqualsCurrentException;
use App\Context\Content\Domain\Model\Level\Exception\TitleEqualsCurrentException;
use App\Context\Content\Domain\Model\Level\Policy\LevelCreatePolicy;
use App\Context\Content\Domain\Model\Level\Policy\LevelEditPolicy;
use App\Context\Content\Domain\Model\Level\Policy\LevelRemovePolicy;
use App\Context\Content\Domain\Model\Level\Policy\LevelRestorePolicy;
use App\Context\Content\Domain\Model\Level\Specification\LevelSlugUniqueSpec;
use App\Context\Content\Domain\Model\Member\MemberId;
use App\Context\Content\Domain\Model\Question\Policy\QuestionCreatePolicy;
use App\Context\Content\Domain\Model\Question\Question;
use App\Context\Content\Domain\Model\Question\QuestionDescription;
use App\Context\Content\Domain\Model\Question\QuestionId;
use App\Context\Content\Domain\Model\Question\QuestionSlug;
use App\Context\Content\Domain\Model\Question\QuestionTitle;
use App\Context\Content\Domain\Model\Question\Specification\QuestionSlugUniqueSpec;
use App\Context\Content\Domain\Model\Skill\SkillId;
use App\Context\Shared\Domain\Aggregate\AggregateRoot;

class Level extends AggregateRoot
{
    private LevelIsRemoved $isRemoved;
    /**
     * @var list<MemberId>
     */
    private array $moderatorList = [];
    private MemberId $ownerId;

    public function __construct(
        LevelCreatePolicy $policy,
        private readonly SkillId $skillId,
        private readonly LevelId $id,
        private LevelTitle $title,
        LevelSlugUniqueSpec $slugSpec,
        private LevelSlug $slug,
        private LevelDescription $description,
    ) {
        parent::__construct();

        if (false === $policy->can($this)) {
            throw new LevelCreatePolicyException($policy);
        }

        $this->ownerId = new MemberId($policy->performer()->id());

        if (false === $slugSpec->isSatisfiedBy($this->skillId, $this->slug)) {
            throw new LevelSlugAlreadyExist($slug);
        }

        $this->record(new LevelCreated(
            $id->value(),
            $this->ownerId->value(),
            $skillId->value(),
            $this->title->value(),
            $this->slug->value(),
            $this->description->value()
        ));

        $this->isRemoved = new LevelIsRemoved(false);
        $this->moderatorList = [];
    }

    public static function create(
        LevelCreatePolicy $policy,
        SkillId $skillId,
        LevelTitle $title,
        LevelSlugUniqueSpec $slugSpec,
        LevelSlug $slug,
        ?LevelDescription $description = null,
        ?LevelId $levelId = null,
    ): self {
        return new self(
            $policy,
            $skillId,
            $levelId ?? LevelId::random(),
            $title,
            $slugSpec,
            $slug,
            $description ?: new LevelDescription('')
        );
    }

    public function getOwnerId(): MemberId
    {
        return $this->ownerId;
    }

    public function getSkillId(): SkillId
    {
        return $this->skillId;
    }

    public function getId(): LevelId
    {
        return $this->id;
    }

    public function getTitle(): LevelTitle
    {
        return $this->title;
    }

    public function getDescription(): ?LevelDescription
    {
        return $this->description;
    }

    public function getSlug(): LevelSlug
    {
        return $this->slug;
    }

    public function getModeratorList(): array
    {
        return $this->moderatorList;
    }

    public function isRemoved(): LevelIsRemoved
    {
        return $this->isRemoved;
    }

    public function assignModerator(LevelEditPolicy $policy, MemberId $memberId): void
    {
        if (false === $policy->can($this)) {
            throw new LevelEditPolicyException($policy);
        }

        if ($this->assertIsModerator($memberId)) {
            throw new LevelModeratorAlreadyAssignedException($this->id, $memberId);
        }

        $this->recordAndApply(
            new LevelModeratorAssigned(
                $this->id->value(),
                $policy->performer()->id(),
                $memberId->value()
            )
        );
    }

    protected function onLevelModeratorAssigned(LevelModeratorAssigned $event): void
    {
        $this->moderatorList[] = new MemberId($event->newModeratorId);
    }

    public function removeModerator(LevelEditPolicy $policy, MemberId $memberId): void
    {
        if (false === $policy->can($this)) {
            throw new LevelEditPolicyException($policy);
        }

        if (!$this->assertIsModerator($memberId)) {
            throw new LevelModeratorNotAssignedException($this->id, $memberId);
        }

        $this->recordAndApply(
            new LevelModeratorRemoved(
                $this->id->value(),
                $policy->performer()->id(),
                $memberId->value(),
            )
        );
    }

    protected function onLevelModeratorRemoved(LevelModeratorRemoved $event): void
    {
        unset($this->moderatorList[array_search($event->moderatorId, $this->moderatorList, true)]);
        $this->moderatorList = array_values($this->moderatorList);
    }

    public function remove(LevelRemovePolicy $policy): void
    {
        if (false === $policy->can($this)) {
            throw new LevelRemovePolicyException($policy);
        }

        if ($this->isRemoved->value()) {
            throw new LevelAlreadyRemoveException($this->id);
        }

        $this->recordAndApply(
            new LevelRemoved(
                $this->id->value(),
                $policy->performer()->id(),
            )
        );
    }

    protected function onLevelRemoved(): void
    {
        $this->isRemoved = new LevelIsRemoved(true);
    }

    public function changeOwner(LevelEditPolicy $policy, MemberId $ownerId): void
    {
        if (false === $policy->can($this)) {
            throw new LevelEditPolicyException($policy);
        }

        if ($this->ownerId->equals($ownerId)) {
            throw new OwnerEqualsCurrentException($this->id, $ownerId);
        }

        $this->recordAndApply(
            new LevelOwnerChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $ownerId->value(),
            )
        );
    }

    protected function onLevelOwnerChanged(LevelOwnerChanged $event): void
    {
        $this->ownerId = new MemberId($event->newOwner);
    }

    public function changeTitle(LevelEditPolicy $policy, LevelTitle $title): void
    {
        if (false === $policy->can($this)) {
            throw new LevelEditPolicyException($policy);
        }

        if ($this->title->equals($title)) {
            throw new TitleEqualsCurrentException($this->id, $title);
        }

        $this->recordAndApply(
            new LevelTitleChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $title->value(),
            )
        );
    }

    public function changeSlug(LevelEditPolicy $policy, LevelSlugUniqueSpec $slugSpec, LevelSlug $slug): void
    {
        if (false === $policy->can($this)) {
            throw new LevelEditPolicyException($policy);
        }

        if ($this->slug->equals($slug)) {
            throw new SlugEqualsCurrentException($this->id, $slug);
        }

        if (false === $slugSpec->isSatisfiedBy($this->skillId, $slug)) {
            throw new LevelSlugAlreadyExist($slug);
        }

        $this->recordAndApply(
            new LevelSlugChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $slug->value(),
            )
        );
    }

    protected function onLevelSlugChanged(LevelSlugChanged $event): void
    {
        $this->slug = new LevelSlug($event->newSlug);
    }

    protected function onLevelTitleChanged(LevelTitleChanged $event): void
    {
        $this->title = new LevelTitle($event->newTitle);
    }

    public function changeDescription(LevelEditPolicy $policy, LevelDescription $description): void
    {
        if (false === $policy->can($this)) {
            throw new LevelEditPolicyException($policy);
        }

        if ($this->description->equals($description)) {
            throw new DescriptionEqualsCurrentException($this->id, $description);
        }

        $this->recordAndApply(
            new LevelDescriptionChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $description->value(),
            )
        );
    }

    protected function onLevelDescriptionChanged(LevelDescriptionChanged $event): void
    {
        $this->description = new LevelDescription($event->newDescription);
    }

    public function createQuestion(
        QuestionCreatePolicy $policy,
        QuestionTitle $questionTitle,
        QuestionSlugUniqueSpec $questionSlugUniqueSpec,
        QuestionSlug $questionSlug,
        ?QuestionDescription $questionDescription = null,
        ?QuestionId $questionId = null,
    ): Question {
        return Question::create(
            $policy,
            $this->getId(),
            $questionTitle,
            $questionSlugUniqueSpec,
            $questionSlug,
            $questionDescription,
            $questionId
        );
    }

    public function assertIsOwner(MemberId $memberId): bool
    {
        return $this->ownerId->equals($memberId);
    }

    public function assertIsModerator(MemberId $memberId): bool
    {
        /** @var MemberId $moderatorId */
        foreach ($this->getModeratorList() as $moderatorId) {
            if ($memberId->equals($moderatorId)) {
                return true;
            }
        }

        return false;
    }

    public function restore(LevelRestorePolicy $policy): void
    {
        if (false === $policy->can($this)) {
            throw new LevelRestorePolicyException($policy);
        }

        if (false === $this->isRemoved()->value()) {
            throw new LevelNotRemovedException($this->id);
        }

        $this->recordAndApply(
            new LevelRestored(
                $this->id->value(),
                $policy->performer()->id(),
            )
        );
    }

    protected function onLevelRestored(): void
    {
        $this->isRemoved = new LevelIsRemoved(false);
    }
}
