<?php

declare(strict_types=1);

namespace App\Context\Content\Domain\Model\Member;

use App\Context\Content\Domain\Model\Member\Event\MemberDeactivated;
use App\Context\Content\Domain\Model\Member\Event\MemberNameChanged;
use App\Context\Content\Domain\Model\Member\Event\MemberRegistered;
use App\Context\Content\Domain\Model\Member\Exception\NameEqualsCurrentException;
use App\Context\Shared\Domain\Aggregate\AggregateRoot;

class Member extends AggregateRoot
{
    public function __construct(
        private readonly MemberId $id,
        private MemberName $name,
    ) {
        parent::__construct();

        $this->record(
            new MemberRegistered(
                $id->value(),
                $this->name->value(),
            )
        );
    }

    public static function create(
        MemberId $id,
        MemberName $name
    ): self {
        return new self(
            $id,
            $name,
        );
    }

    public function getId(): MemberId
    {
        return $this->id;
    }

    public function getName(): MemberName
    {
        return $this->name;
    }

    public function deactivate(): void
    {
        $this->record(
            new MemberDeactivated(
                $this->id->value(),
            )
        );
    }

    public function changeName(MemberName $name): void
    {
        if ($this->name->equals($name)) {
            throw new NameEqualsCurrentException($this->id, $name);
        }

        $this->recordAndApply(
            new MemberNameChanged(
                $this->id->value(),
                $name->value(),
            )
        );
    }

    protected function onMemberNameChanged(MemberNameChanged $event): void
    {
        $this->name = new MemberName($event->newName);
    }
}
