<?php

declare(strict_types=1);

namespace App\Context\Content\Domain\Model\Question;

use App\Context\Content\Domain\Model\Level\LevelId;
use App\Context\Content\Domain\Model\Member\MemberId;
use App\Context\Content\Domain\Model\Question\Event\QuestionCreated;
use App\Context\Content\Domain\Model\Question\Event\QuestionDescriptionChanged;
use App\Context\Content\Domain\Model\Question\Event\QuestionRemoved;
use App\Context\Content\Domain\Model\Question\Event\QuestionRestored;
use App\Context\Content\Domain\Model\Question\Event\QuestionSlugChanged;
use App\Context\Content\Domain\Model\Question\Event\QuestionTitleChanged;
use App\Context\Content\Domain\Model\Question\Exception\DescriptionEqualsCurrentException;
use App\Context\Content\Domain\Model\Question\Exception\Policy\QuestionCreatePolicyException;
use App\Context\Content\Domain\Model\Question\Exception\Policy\QuestionEditPolicyException;
use App\Context\Content\Domain\Model\Question\Exception\Policy\QuestionRemovePolicyException;
use App\Context\Content\Domain\Model\Question\Exception\Policy\QuestionRestorePolicyException;
use App\Context\Content\Domain\Model\Question\Exception\QuestionAlreadyRemoveException;
use App\Context\Content\Domain\Model\Question\Exception\QuestionNotRemovedException;
use App\Context\Content\Domain\Model\Question\Exception\QuestionSlugAlreadyExist;
use App\Context\Content\Domain\Model\Question\Exception\SlugEqualsCurrentException;
use App\Context\Content\Domain\Model\Question\Exception\TitleEqualsCurrentException;
use App\Context\Content\Domain\Model\Question\Policy\QuestionCreatePolicy;
use App\Context\Content\Domain\Model\Question\Policy\QuestionEditPolicy;
use App\Context\Content\Domain\Model\Question\Policy\QuestionRemovePolicy;
use App\Context\Content\Domain\Model\Question\Policy\QuestionRestorePolicy;
use App\Context\Content\Domain\Model\Question\Specification\QuestionSlugUniqueSpec;
use App\Context\Shared\Domain\Aggregate\AggregateRoot;

class Question extends AggregateRoot
{
    private QuestionIsRemoved $isRemoved;
    private MemberId $ownerId;

    public function __construct(
        QuestionCreatePolicy $policy,
        private readonly LevelId $levelId,
        private readonly QuestionId $id,
        private QuestionTitle $title,
        QuestionSlugUniqueSpec $slugSpec,
        private QuestionSlug $slug,
        private QuestionDescription $description,
    ) {
        parent::__construct();

        if (false === $policy->can($this)) {
            throw new QuestionCreatePolicyException($policy);
        }

        $this->ownerId = new MemberId($policy->performer()->id());

        if (false === $slugSpec->isSatisfiedBy($this->levelId, $slug)) {
            throw new QuestionSlugAlreadyExist($slug);
        }

        $this->record(
            new QuestionCreated(
                $id->value(),
                $this->ownerId->value(),
                $this->levelId->value(),
                $this->title->value(),
                $this->slug->value(),
                $this->description->value(),
            )
        );

        $this->isRemoved = new QuestionIsRemoved(false);
    }

    public static function create(
        QuestionCreatePolicy $policy,
        LevelId $levelId,
        QuestionTitle $title,
        QuestionSlugUniqueSpec $slugSpec,
        QuestionSlug $slug,
        ?QuestionDescription $description = null,
        ?QuestionId $questionId = null
    ): self {
        return new self(
            $policy,
            $levelId,
            $questionId ?? QuestionId::random(),
            $title,
            $slugSpec,
            $slug,
            $description ?: new QuestionDescription(''),
        );
    }

    public function getLevelId(): LevelId
    {
        return $this->levelId;
    }

    public function getId(): QuestionId
    {
        return $this->id;
    }

    public function getTitle(): QuestionTitle
    {
        return $this->title;
    }

    public function getDescription(): ?QuestionDescription
    {
        return $this->description;
    }

    public function getSlug(): QuestionSlug
    {
        return $this->slug;
    }

    public function isRemoved(): QuestionIsRemoved
    {
        return $this->isRemoved;
    }

    public function remove(QuestionRemovePolicy $policy): void
    {
        if (false === $policy->can($this)) {
            throw new QuestionRemovePolicyException($policy);
        }

        if ($this->isRemoved->value()) {
            throw new QuestionAlreadyRemoveException($this->id);
        }

        $this->recordAndApply(
            new QuestionRemoved(
                $this->id->value(),
                $policy->performer()->id(),
            )
        );
    }

    protected function onQuestionRemoved(): void
    {
        $this->isRemoved = new QuestionIsRemoved(true);
    }

    public function changeTitle(QuestionEditPolicy $policy, QuestionTitle $title): void
    {
        if (false === $policy->can($this)) {
            throw new QuestionEditPolicyException($policy);
        }

        if ($this->title->equals($title)) {
            throw new TitleEqualsCurrentException($this->id, $title);
        }

        $this->recordAndApply(
            new QuestionTitleChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $title->value(),
            )
        );
    }

    protected function onQuestionTitleChanged(QuestionTitleChanged $event): void
    {
        $this->title = new QuestionTitle($event->title);
    }

    public function changeSlug(QuestionEditPolicy $policy, QuestionSlugUniqueSpec $slugSpec, QuestionSlug $slug): void
    {
        if (false === $policy->can($this)) {
            throw new QuestionEditPolicyException($policy);
        }

        if ($this->slug->equals($slug)) {
            throw new SlugEqualsCurrentException($this->id, $slug);
        }

        if (false === $slugSpec->isSatisfiedBy($this->levelId, $slug)) {
            throw new QuestionSlugAlreadyExist($slug);
        }

        $this->recordAndApply(
            new QuestionSlugChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $slug->value(),
            )
        );
    }

    protected function onQuestionSlugChanged(QuestionSlugChanged $event): void
    {
        $this->slug = new QuestionSlug($event->newSlug);
    }

    public function changeDescription(QuestionEditPolicy $policy, QuestionDescription $description): void
    {
        if (false === $policy->can($this)) {
            throw new Exception\Policy\QuestionEditPolicyException($policy);
        }

        if ($this->description->equals($description)) {
            throw new DescriptionEqualsCurrentException($this->id, $description);
        }

        $this->recordAndApply(
            new QuestionDescriptionChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $description->value(),
            )
        );
    }

    protected function onQuestionDescriptionChanged(QuestionDescriptionChanged $event): void
    {
        $this->description = new QuestionDescription($event->description);
    }

    public function restore(QuestionRestorePolicy $policy): void
    {
        if (false === $policy->can($this)) {
            throw new QuestionRestorePolicyException($policy);
        }

        if (false === $this->isRemoved->value()) {
            throw new QuestionNotRemovedException($this->id);
        }

        $this->recordAndApply(
            new QuestionRestored(
                $this->id->value(),
                $policy->performer()->id(),
            )
        );
    }

    protected function onQuestionRestored(): void
    {
        $this->isRemoved = new QuestionIsRemoved(false);
    }
}
