<?php

declare(strict_types=1);

namespace App\Context\Content\Domain\Model\Skill;

use App\Context\Content\Domain\Model\Level\Level;
use App\Context\Content\Domain\Model\Level\LevelDescription;
use App\Context\Content\Domain\Model\Level\LevelId;
use App\Context\Content\Domain\Model\Level\LevelSlug;
use App\Context\Content\Domain\Model\Level\LevelTitle;
use App\Context\Content\Domain\Model\Level\Policy\LevelCreatePolicy;
use App\Context\Content\Domain\Model\Level\Specification\LevelSlugUniqueSpec;
use App\Context\Content\Domain\Model\Member\MemberId;
use App\Context\Content\Domain\Model\Skill\Event\SkillCreated;
use App\Context\Content\Domain\Model\Skill\Event\SkillDescriptionChanged;
use App\Context\Content\Domain\Model\Skill\Event\SkillModeratorAssigned;
use App\Context\Content\Domain\Model\Skill\Event\SkillModeratorRemoved;
use App\Context\Content\Domain\Model\Skill\Event\SkillOwnerChanged;
use App\Context\Content\Domain\Model\Skill\Event\SkillRemoved;
use App\Context\Content\Domain\Model\Skill\Event\SkillRestored;
use App\Context\Content\Domain\Model\Skill\Event\SkillSlugChanged;
use App\Context\Content\Domain\Model\Skill\Event\SkillTitleChanged;
use App\Context\Content\Domain\Model\Skill\Exception\DescriptionEqualsCurrentException;
use App\Context\Content\Domain\Model\Skill\Exception\OwnerEqualsCurrentException;
use App\Context\Content\Domain\Model\Skill\Exception\Policy\SkillCreatePolicyException;
use App\Context\Content\Domain\Model\Skill\Exception\Policy\SkillEditPolicyException;
use App\Context\Content\Domain\Model\Skill\Exception\Policy\SkillRemovePolicyException;
use App\Context\Content\Domain\Model\Skill\Exception\Policy\SkillRestorePolicyException;
use App\Context\Content\Domain\Model\Skill\Exception\SkillAlreadyRemoveException;
use App\Context\Content\Domain\Model\Skill\Exception\SkillModeratorAlreadyAssignedException;
use App\Context\Content\Domain\Model\Skill\Exception\SkillModeratorNotAssignedException;
use App\Context\Content\Domain\Model\Skill\Exception\SkillNotRemovedException;
use App\Context\Content\Domain\Model\Skill\Exception\SkillSlugAlreadyExist;
use App\Context\Content\Domain\Model\Skill\Exception\SlugEqualsCurrentException;
use App\Context\Content\Domain\Model\Skill\Exception\TitleEqualsCurrentException;
use App\Context\Content\Domain\Model\Skill\Policy\SkillCreatePolicy;
use App\Context\Content\Domain\Model\Skill\Policy\SkillEditPolicy;
use App\Context\Content\Domain\Model\Skill\Policy\SkillRemovePolicy;
use App\Context\Content\Domain\Model\Skill\Policy\SkillRestorePolicy;
use App\Context\Content\Domain\Model\Skill\Specification\SkillSlugUniqueSpec;
use App\Context\Shared\Domain\Aggregate\AggregateRoot;

class Skill extends AggregateRoot
{
    private SkillIsRemoved $isRemoved;
    /**
     * @var list<MemberId>
     */
    private array $moderatorList = [];
    private MemberId $ownerId;

    public function __construct(
        SkillCreatePolicy $policy,
        private readonly SkillId $id,
        private SkillTitle $title,
        SkillSlugUniqueSpec $slugSpec,
        private SkillSlug $slug,
        private SkillDescription $description,
    ) {
        parent::__construct();

        if (false === $policy->can($this)) {
            throw new SkillCreatePolicyException($policy);
        }

        $this->ownerId = new MemberId($policy->performer()->id());

        if (false === $slugSpec->isSatisfiedBy($slug)) {
            throw new SkillSlugAlreadyExist($slug);
        }

        $this->record(
            new SkillCreated(
                $id->value(),
                $this->ownerId->value(),
                $this->title->value(),
                $this->slug->value(),
                $this->description->value()
            )
        );

        $this->isRemoved = new SkillIsRemoved(false);
        $this->moderatorList = [];
    }

    public static function create(
        SkillCreatePolicy $policy,
        SkillTitle $title,
        SkillSlugUniqueSpec $slugSpec,
        SkillSlug $slug,
        ?SkillDescription $description = null,
        SkillId $id = null
    ): self {
        return new self(
            $policy,
            $id ?? SkillId::random(),
            $title,
            $slugSpec,
            $slug,
            $description ?: new SkillDescription(''),
        );
    }

    public function getOwnerId(): MemberId
    {
        return $this->ownerId;
    }

    public function getId(): SkillId
    {
        return $this->id;
    }

    public function getTitle(): SkillTitle
    {
        return $this->title;
    }

    public function getSlug(): SkillSlug
    {
        return $this->slug;
    }

    public function getDescription(): ?SkillDescription
    {
        return $this->description;
    }

    public function getModeratorList(): array
    {
        return $this->moderatorList;
    }

    public function isRemoved(): SkillIsRemoved
    {
        return $this->isRemoved;
    }

    public function assignModerator(SkillEditPolicy $policy, MemberId $memberId): void
    {
        if (false === $policy->can($this)) {
            throw new SkillEditPolicyException($policy);
        }

        if ($this->assertIsModerator($memberId)) {
            throw new SkillModeratorAlreadyAssignedException($this->id, $memberId);
        }

        $this->recordAndApply(
            new SkillModeratorAssigned(
                $this->id->value(),
                $policy->performer()->id(),
                $memberId->value()
            )
        );
    }

    protected function onSkillModeratorAssigned(SkillModeratorAssigned $event): void
    {
        $this->moderatorList[] = new MemberId($event->newModeratorId);
    }

    public function removeModerator(SkillEditPolicy $policy, MemberId $memberId): void
    {
        if (false === $policy->can($this)) {
            throw new SkillEditPolicyException($policy);
        }

        if (!$this->assertIsModerator($memberId)) {
            throw new SkillModeratorNotAssignedException($this->id, $memberId);
        }

        $this->recordAndApply(
            new SkillModeratorRemoved(
                $this->id->value(),
                $policy->performer()->id(),
                $memberId->value(),
            )
        );
    }

    protected function onSkillModeratorRemoved(SkillModeratorRemoved $event): void
    {
        unset($this->moderatorList[array_search($event->moderatorId, $this->moderatorList, true)]);
        $this->moderatorList = array_values($this->moderatorList);
    }

    public function remove(SkillRemovePolicy $policy): void
    {
        if (false === $policy->can($this)) {
            throw new SkillRemovePolicyException($policy);
        }

        if ($this->isRemoved->value()) {
            throw new SkillAlreadyRemoveException($this->id);
        }

        $this->recordAndApply(
            new SkillRemoved(
                $this->id->value(),
                $policy->performer()->id(),
            )
        );
    }

    protected function onSkillRemoved(): void
    {
        $this->isRemoved = new SkillIsRemoved(true);
    }

    public function changeOwner(SkillEditPolicy $policy, MemberId $ownerId): void
    {
        if (false === $policy->can($this)) {
            throw new SkillEditPolicyException($policy);
        }

        if ($this->ownerId->equals($ownerId)) {
            throw new OwnerEqualsCurrentException($this->id, $ownerId);
        }

        $this->recordAndApply(
            new SkillOwnerChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $ownerId->value(),
            )
        );
    }

    protected function onSkillOwnerChanged(SkillOwnerChanged $event): void
    {
        $this->ownerId = new MemberId($event->newOwner);
    }

    public function changeTitle(SkillEditPolicy $policy, SkillTitle $title): void
    {
        if (false === $policy->can($this)) {
            throw new SkillEditPolicyException($policy);
        }

        if ($this->title->equals($title)) {
            throw new TitleEqualsCurrentException($this->id, $title);
        }

        $this->recordAndApply(
            new SkillTitleChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $title->value(),
            )
        );
    }

    protected function onSkillTitleChanged(SkillTitleChanged $event): void
    {
        $this->title = new SkillTitle($event->newTitle);
    }

    public function changeSlug(SkillEditPolicy $policy, SkillSlugUniqueSpec $slugSpec, SkillSlug $slug): void
    {
        if (false === $policy->can($this)) {
            throw new SkillEditPolicyException($policy);
        }

        if ($this->slug->equals($slug)) {
            throw new SlugEqualsCurrentException($this->id, $slug);
        }

        if (false === $slugSpec->isSatisfiedBy($slug)) {
            throw new SkillSlugAlreadyExist($slug);
        }

        $this->recordAndApply(
            new SkillSlugChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $slug->value(),
            )
        );
    }

    protected function onSkillSlugChanged(SkillSlugChanged $event): void
    {
        $this->slug = new SkillSlug($event->newSlug);
    }

    public function changeDescription(SkillEditPolicy $policy, SkillDescription $description): void
    {
        if (false === $policy->can($this)) {
            throw new SkillEditPolicyException($policy);
        }

        if ($this->description->equals($description)) {
            throw new DescriptionEqualsCurrentException($this->id, $description);
        }

        $this->recordAndApply(
            new SkillDescriptionChanged(
                $this->id->value(),
                $policy->performer()->id(),
                $description->value(),
            )
        );
    }

    protected function onSkillDescriptionChanged(SkillDescriptionChanged $event): void
    {
        $this->description = new SkillDescription($event->newDescription);
    }

    public function createLevel(
        LevelCreatePolicy $policy,
        LevelTitle $levelTitle,
        LevelSlugUniqueSpec $levelSlugUniqueSpec,
        LevelSlug $levelSlug,
        ?LevelDescription $levelDescription = null,
        ?LevelId $levelId = null,
    ): Level {
        return Level::create(
            $policy,
            $this->getId(),
            $levelTitle,
            $levelSlugUniqueSpec,
            $levelSlug,
            $levelDescription,
            $levelId
        );
    }

    public function assertIsOwner(MemberId $memberId): bool
    {
        return $this->ownerId->equals($memberId);
    }

    public function assertIsModerator(MemberId $memberId): bool
    {
        /** @var MemberId $moderatorId */
        foreach ($this->getModeratorList() as $moderatorId) {
            if ($memberId->equals($moderatorId)) {
                return true;
            }
        }

        return false;
    }

    public function restore(SkillRestorePolicy $policy): void
    {
        if (false === $policy->can($this)) {
            throw new SkillRestorePolicyException($policy);
        }

        if ($this->isRemoved->value()) {
            throw new SkillNotRemovedException($this->id);
        }

        $this->recordAndApply(
            new SkillRestored(
                $this->id->value(),
                $policy->performer()->id(),
            )
        );
    }

    protected function onSkillRestored(): void
    {
        $this->isRemoved = new SkillIsRemoved(false);
    }
}
