<?php

declare(strict_types=1);

namespace App\Context\Content\Infrastructure\Import\Loader;

use App\Context\Content\Application\Command\Level\CreateLevel\CreateLevelCommand;
use App\Context\Content\Application\Command\Level\RemoveLevel\RemoveLevelCommand;
use App\Context\Content\Application\Command\Level\RestoreLevel\RestoreLevelCommand;
use App\Context\Content\Application\Command\Level\UpdateLevel\UpdateLevelCommand;
use App\Context\Content\Application\Process\Import\Content\Dto\Level;
use App\Context\Content\Application\Process\Import\Content\Loader\LevelLoader;
use App\Context\Content\Domain\Model\Level\Exception\LevelAlreadyRemoveException;
use App\Context\Content\Domain\Model\Level\Exception\LevelNotFoundException;
use App\Context\Content\Domain\Model\Level\Level as DomainLevel;
use App\Context\Content\Domain\Model\Level\LevelId;
use App\Context\Content\Domain\Model\Level\LevelRepositoryReader;
use App\Context\Content\Domain\Model\Skill\Exception\SkillNotFoundException;
use App\Context\Shared\Domain\Bus\Command\CommandBus;

class LevelBusLoader implements LevelLoader
{
    public function __construct(
        private readonly LevelRepositoryReader $levelRepositoryReader,
        private readonly CommandBus $commandBus,
    ) {
    }

    public function load(Level $level): void
    {
        try {
            $domainLevel = $this->levelRepositoryReader->findById(new LevelId($level->id));
            $this->update($level, $domainLevel);
        } catch (LevelNotFoundException) {
            $this->create($level);
        }
    }

    protected function create(Level $level): void
    {
        $this->commandBus->dispatch(
            new CreateLevelCommand(
                $level->skillId,
                $this->prepareTitle($level),
                $level->slug,
                $level->description,
                $level->id,
            )
        );
    }

    protected function update(Level $level, DomainLevel $domainLevel): void
    {
        if ($domainLevel->isRemoved()->value()) {
            $this->commandBus->dispatch(
                new RestoreLevelCommand(
                    $level->skillId,
                    $level->id,
                )
            );
        }

        $this->commandBus->dispatch(
            new UpdateLevelCommand(
                $level->skillId,
                $level->id,
                $this->prepareTitle($level),
                $level->slug,
                $level->description
            )
        );
    }

    protected function prepareTitle(Level $level): string
    {
        return sprintf(
            '%s. %s',
            str_pad((string) $level->sort, 2, '0', STR_PAD_LEFT),
            $level->title
        );
    }

    public function remove(string $skillId, string $levelId): void
    {
        try {
            $this->commandBus->dispatch(
                new RemoveLevelCommand(
                    $skillId,
                    $levelId,
                )
            );
        } catch (\Exception $e) {
            $pe = $e->getPrevious();

            if ($e instanceof LevelAlreadyRemoveException || $pe instanceof LevelAlreadyRemoveException) {
                return;
            }

            if ($e instanceof LevelNotFoundException || $pe instanceof LevelNotFoundException) {
                return;
            }

            if ($e instanceof SkillNotFoundException || $pe instanceof SkillNotFoundException) {
                return;
            }

            throw $e;
        }
    }
}
