<?php

declare(strict_types=1);

namespace App\Context\Content\Infrastructure\Import\Loader;

use App\Context\Content\Application\Command\Question\CreateQuestion\CreateQuestionCommand;
use App\Context\Content\Application\Command\Question\RemoveQuestion\RemoveQuestionCommand;
use App\Context\Content\Application\Command\Question\RestoreQuestion\RestoreQuestionCommand;
use App\Context\Content\Application\Command\Question\UpdateQuestion\UpdateQuestionCommand;
use App\Context\Content\Application\Process\Import\Content\Dto\Question;
use App\Context\Content\Application\Process\Import\Content\Loader\QuestionLoader;
use App\Context\Content\Domain\Model\Level\Exception\LevelNotFoundException;
use App\Context\Content\Domain\Model\Question\Exception\QuestionAlreadyRemoveException;
use App\Context\Content\Domain\Model\Question\Exception\QuestionNotFoundException;
use App\Context\Content\Domain\Model\Question\Question as DomainQuestion;
use App\Context\Content\Domain\Model\Question\QuestionId;
use App\Context\Content\Domain\Model\Question\QuestionRepositoryReader;
use App\Context\Content\Domain\Model\Skill\Exception\SkillNotFoundException;
use App\Context\Shared\Domain\Bus\Command\CommandBus;

class QuestionBusLoader implements QuestionLoader
{
    public function __construct(
        private readonly QuestionRepositoryReader $repository,
        private readonly CommandBus $commandBus,
    ) {
    }

    public function load(Question $question): void
    {
        try {
            $domainQuestion = $this->repository->findById(new QuestionId($question->id));
            $this->update($question, $domainQuestion);
        } catch (QuestionNotFoundException) {
            $this->create($question);
        }
    }

    protected function update(Question $question, DomainQuestion $domainQuestion): void
    {
        if ($domainQuestion->isRemoved()->value()) {
            $this->commandBus->dispatch(
                new RestoreQuestionCommand(
                    $question->skillId,
                    $question->levelId,
                    $question->id,
                )
            );
        }

        $this->commandBus->dispatch(
            new UpdateQuestionCommand(
                $question->skillId,
                $question->levelId,
                $question->id,
                $this->prepareTitle($question),
                $question->slug,
                $question->description,
            )
        );
    }

    protected function create(Question $question): void
    {
        $this->commandBus->dispatch(
            new CreateQuestionCommand(
                $question->skillId,
                $question->levelId,
                $this->prepareTitle($question),
                $question->slug,
                $question->description,
                $question->id,
            )
        );
    }

    protected function prepareTitle(Question $question): string
    {
        return sprintf(
            '%s. %s',
            str_pad((string) $question->sort, 2, '0', STR_PAD_LEFT),
            $question->title
        );
    }

    public function remove(string $skillId, string $levelId, string $questionId): void
    {
        try {
            $this->commandBus->dispatch(
                new RemoveQuestionCommand(
                    $skillId,
                    $levelId,
                    $questionId
                )
            );
        } catch (\Exception $e) {
            $pe = $e->getPrevious();

            if ($e instanceof QuestionAlreadyRemoveException || $pe instanceof QuestionAlreadyRemoveException) {
                return;
            }

            if ($e instanceof QuestionNotFoundException || $pe instanceof QuestionNotFoundException) {
                return;
            }

            if ($e instanceof LevelNotFoundException || $pe instanceof LevelNotFoundException) {
                return;
            }

            if ($e instanceof SkillNotFoundException || $pe instanceof SkillNotFoundException) {
                return;
            }

            throw $e;
        }
    }
}
