<?php

declare(strict_types=1);

namespace App\Context\Content\Infrastructure\Import\Loader;

use App\Context\Content\Application\Command\Skill\CreateSkill\CreateSkillCommand;
use App\Context\Content\Application\Command\Skill\PatchSkill\PatchSkillCommand;
use App\Context\Content\Application\Command\Skill\RemoveSkill\RemoveSkillCommand;
use App\Context\Content\Application\Command\Skill\RestoreSkill\RestoreSkillCommand;
use App\Context\Content\Application\Process\Import\Content\Dto\Skill;
use App\Context\Content\Application\Process\Import\Content\Loader\SkillLoader;
use App\Context\Content\Domain\Model\Skill\Exception\SkillAlreadyRemoveException;
use App\Context\Content\Domain\Model\Skill\Exception\SkillNotFoundException;
use App\Context\Content\Domain\Model\Skill\Skill as SkillDomain;
use App\Context\Content\Domain\Model\Skill\SkillId;
use App\Context\Content\Domain\Model\Skill\SkillRepositoryReader;
use App\Context\Shared\Domain\Bus\Command\CommandBus;

class SkillBusLoader implements SkillLoader
{
    public function __construct(
        private readonly SkillRepositoryReader $repositoryReader,
        private readonly CommandBus $commandBus,
    ) {
    }

    public function load(Skill $skill): void
    {
        try {
            $skillDomain = $this->repositoryReader->findById(new SkillId($skill->id));
            $this->update($skill, $skillDomain);
        } catch (SkillNotFoundException) {
            $this->create($skill);
        }
    }

    protected function create(Skill $skill): void
    {
        $this->commandBus->dispatch(
            new CreateSkillCommand(
                $this->prepareTitle($skill),
                $skill->slug,
                $skill->description,
                $skill->id
            )
        );
    }

    protected function update(Skill $skill, SkillDomain $skillDomain): void
    {
        if ($skillDomain->isRemoved()->value()) {
            $this->commandBus->dispatch(
                new RestoreSkillCommand(
                    $skill->id,
                )
            );
        }

        $this->commandBus->dispatch(
            new PatchSkillCommand(
                $skill->id,
                $this->prepareTitle($skill),
                $skill->slug,
                $skill->description,
            )
        );
    }

    protected function prepareTitle(Skill $skill): string
    {
        return sprintf(
            '%s. %s',
            str_pad((string) $skill->sort, 2, '0', STR_PAD_LEFT),
            $skill->title
        );
    }

    public function remove(string $id): void
    {
        try {
            $this->commandBus->dispatch(
                new RemoveSkillCommand($id)
            );
        } catch (\Exception $e) {
            $pe = $e->getPrevious();

            if ($e instanceof SkillNotFoundException || $pe instanceof SkillNotFoundException) {
                return;
            }

            if ($e instanceof SkillAlreadyRemoveException || $pe instanceof SkillAlreadyRemoveException) {
                return;
            }

            throw $e;
        }
    }
}
