<?php

declare(strict_types=1);

namespace App\Context\Content\Infrastructure\Import\Yaml;

use App\Context\Content\Application\Process\Import\Content\Dto\Level;
use App\Context\Content\Application\Process\Import\Content\Dto\Question;
use App\Context\Content\Application\Process\Import\Content\Dto\Skill;
use App\Context\Content\Application\Process\Import\Content\ImportService;
use App\Context\Content\Application\Process\Import\Content\Loader\LevelLoader;
use App\Context\Content\Application\Process\Import\Content\Loader\QuestionLoader;
use App\Context\Content\Application\Process\Import\Content\Loader\SkillLoader;
use Doctrine\DBAL\Query\QueryBuilder;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Finder\Finder;
use Symfony\Component\Yaml\Yaml;

class ImportYaml implements ImportService
{
    private array $skillIds = [];
    private array $levelIds = [];
    private array $questionIds = [];

    public function __construct(
        private readonly string $root,
        private readonly SkillLoader $skillLoader,
        private readonly LevelLoader $levelLoader,
        private readonly QuestionLoader $questionLoader,
        private readonly EntityManagerInterface $em,
    ) {
    }

    public function import(): void
    {
        $finder = (new Finder())
            ->in($this->root)
            ->name('*.yaml')
            ->name('*.yml')
            ->contains('/skills:/i')
            ->files();

        foreach ($finder as $file) {
            $this->importSkills($file->getPathname());
        }

        $this->remove();
    }

    protected function importSkills(string $filename): void
    {
        $content = file_get_contents($filename);
        $currentDir = dirname($filename);

        /** @var array $rawSkills */
        $rawSkills = Yaml::parse($content);

        if (!isset($rawSkills['skills']) || !is_array($rawSkills['skills'])) {
            return;
        }

        /**
         * @var array<int, array> $rawSkills
         */
        $rawSkills = $rawSkills['skills'];

        foreach ($rawSkills as $key => $rawSkill) {
            $skill = new Skill(
                (string) ($rawSkill['id'] ?? ''),
                (string) ($rawSkill['slug'] ?? ''),
                (string) ($rawSkill['title'] ?? ''),
                (int) ($rawSkill['sort'] ?? $key + 1),
                (string) ($rawSkill['description'] ?? 0),
            );

            $this->skillLoader->load($skill);

            if (isset($rawSkill['level_files']) && is_array($rawSkill['level_files'])) {
                /** @var array<int, string> $levelFiles */
                $levelFiles = array_filter($rawSkill['level_files'], fn (mixed $value) => is_string($value));
                $this->importLevel($skill->id, $levelFiles, $currentDir);
            }

            $this->skillIds[] = $skill->id;
        }
    }

    /**
     * @param array<int, string> $paths
     */
    protected function importLevel(string $skillId, array $paths, string $root): void
    {
        foreach ($paths as $key => $levelPath) {
            $path = $root.'/'.$levelPath;
            $content = file_get_contents($path);

            /** @var array $rawLevels */
            $rawLevels = Yaml::parse($content);

            if (!isset($rawLevels['levels']) || !is_array($rawLevels['levels'])) {
                continue;
            }

            /**
             * @var array<int, array> $rawLevels
             */
            $rawLevels = $rawLevels['levels'];

            foreach ($rawLevels as $rawLevel) {
                $level = new Level(
                    $skillId,
                    (string) ($rawLevel['id'] ?? ''),
                    (string) ($rawLevel['slug'] ?? ''),
                    (string) ($rawLevel['title'] ?? ''),
                    (int) ($rawLevel['sort'] ?? ($key + 1)),
                    (string) ($rawLevel['description'] ?? ''),
                );

                $this->levelLoader->load($level);

                if (isset($rawLevel['questions']) && is_array($rawLevel['questions'])) {
                    /** @var array<int, array> $questions */
                    $questions = array_filter($rawLevel['questions'], fn (mixed $value) => is_array($value));
                    $this->importQuestions($skillId, $level->id, $questions);
                }

                $this->levelIds[] = $level->id;
            }
        }
    }

    /**
     * @param array<int, array> $questions
     */
    protected function importQuestions(string $skillId, string $levelId, array $questions): void
    {
        foreach ($questions as $key => $question) {
            $question = new Question(
                $skillId,
                $levelId,
                (string) ($question['id'] ?? ''),
                (string) ($question['slug'] ?? ''),
                (string) ($question['title'] ?? ''),
                (int) ($question['sort'] ?? ($key + 1)),
                (string) ($question['description'] ?? ''),
            );

            $this->questionLoader->load($question);
            $this->questionIds[] = $question->id;
        }
    }

    public function remove(): void
    {
        $qb = new QueryBuilder($this->em->getConnection());
        $questionsIdsForDeleting = $qb->from('question', 'q')
            ->select('q.id question', 'l.id level', 's.id skill')
            ->innerJoin('q', 'level', 'l', 'q.level_id = l.id')
            ->innerJoin('l', 'skill', 's', 's.id = l.skill_id')
            ->where(
                'q.id::text NOT IN (:questions_ids) AND q.is_removed = false AND s.is_removed = false AND l.is_removed = false'
            )
            ->setParameter(
                'questions_ids',
                $this->questionIds,
                \Doctrine\DBAL\Connection::PARAM_STR_ARRAY
            )
            ->fetchAllAssociative();

        if (false === empty($questionsIdsForDeleting)) {
            foreach ($questionsIdsForDeleting as $data) {
                /** @var array{skill: string, level: string, question: string} $data */
                $this->questionLoader->remove($data['skill'], $data['level'], $data['question']);
            }
        }

        $qb = new QueryBuilder($this->em->getConnection());
        $levelIdsForDeleting = $qb->from('level', 'l')
            ->select('l.id level', 's.id skill')
            ->innerJoin('l', 'skill', 's', 'l.skill_id = s.id')
            ->where('l.id::text NOT IN (:level_ids) AND l.is_removed = false AND s.is_removed = false')
            ->setParameter(
                'level_ids',
                $this->levelIds,
                \Doctrine\DBAL\Connection::PARAM_STR_ARRAY
            )
            ->fetchAllAssociative();

        if (false === empty($levelIdsForDeleting)) {
            foreach ($levelIdsForDeleting as $data) {
                /** @var array{skill: string, level: string} $data */
                $this->levelLoader->remove($data['skill'], $data['level']);
            }
        }

        $qb = new QueryBuilder($this->em->getConnection());
        $skillIdsForDeleting = $qb->from('skill', 's')
            ->select('s.id')
            ->where('s.id::text NOT IN (:skill_ids) AND s.is_removed = false')
            ->setParameter(
                'skill_ids',
                $this->skillIds,
                \Doctrine\DBAL\Connection::PARAM_STR_ARRAY
            )
            ->fetchAllAssociative();

        if (false === empty($skillIdsForDeleting)) {
            foreach ($skillIdsForDeleting as $data) {
                /** @var array{id: string} $data */
                $this->skillLoader->remove($data['id']);
            }
        }
    }
}
