<?php

declare(strict_types=1);

namespace App\Context\Content\Infrastructure\Repository\Application\Process\Export\Content\Dbal;

use App\Context\Content\Application\Process\Export\Content\Dto\Level;
use App\Context\Content\Application\Process\Export\Content\Dto\Question;
use App\Context\Content\Application\Process\Export\Content\LevelProvider;
use Doctrine\DBAL\Connection;

class DbalLevelProvider implements LevelProvider
{
    public function __construct(
        private readonly Connection $connection,
        private readonly int $batchSize = 100,
    ) {
    }

    public function get(array $skillIds): \Iterator
    {
        $offset = 0;
        while (true) {
            /** @var list<array> $rawLevels */
            $rawLevels = $this->connection->fetchAllAssociative(
                'SELECT id, slug, title, description'
                .' FROM level l'
                .' WHERE l.is_removed = false '
                .' AND l.skill_id::text IN ( ? )'
                .' ORDER BY title'
                .' LIMIT ? '
                .' OFFSET ? ',
                [
                    $skillIds,
                    $this->batchSize,
                    $offset,
                ],
                [
                    Connection::PARAM_STR_ARRAY,
                ]
            );

            if (empty($rawLevels)) {
                break;
            }

            /** @var string[] $ids */
            $ids = array_column($rawLevels, 'id');

            $questions = $this->getQuestionsForLevels($ids);

            foreach ($rawLevels as $rawLevel) {
                if (false === isset($rawLevel['id'])) {
                    continue;
                }

                $levelQuestions = $questions[(string) $rawLevel['id']] ?? [];

                yield $this->toObject($rawLevel, $levelQuestions);
            }

            $offset += $this->batchSize;
        }
    }

    /**
     * @param array<Question> $questions
     */
    private function toObject(array $data, array $questions): Level
    {
        return new Level(
            (string) ($data['id'] ?? ''),
            (string) ($data['slug'] ?? ''),
            (string) ($data['title'] ?? ''),
            (string) ($data['description'] ?? ''),
            $questions
        );
    }

    /**
     * @param string[] $levelIds
     *
     * @return array<string, array<Question>>
     *
     * @throws \Doctrine\DBAL\Exception
     */
    private function getQuestionsForLevels(array $levelIds): array
    {
        $rawQuestions = $this->connection->fetchAllAssociative(
            'SELECT id, slug, title, description, level_id'
            .' FROM question q'
            .' WHERE q.is_removed = false '
            .' AND q.level_id::text IN ( ? )'
            .' ORDER BY title',
            [
                $levelIds,
            ],
            [
                Connection::PARAM_STR_ARRAY,
            ]
        );

        $questions = [];
        foreach ($rawQuestions as $rawQuestion) {
            $questions[(string) $rawQuestion['level_id']][] = new Question(
                (string) ($rawQuestion['id'] ?? ''),
                (string) ($rawQuestion['slug'] ?? ''),
                (string) ($rawQuestion['title'] ?? ''),
                (string) ($rawQuestion['description'] ?? ''),
            );
        }

        return $questions;
    }
}
