<?php

declare(strict_types=1);

namespace App\Context\Content\Infrastructure\Repository\Application\Query\Level\Doctrine;

use App\Context\Content\Application\Query\Level\ReadModel\LevelNotFoundBySlugException;
use App\Context\Content\Application\Query\Level\ReadModel\LevelReadModel;
use App\Context\Content\Application\Query\Level\ReadModel\LevelReadRepository;
use App\Context\Content\Domain\Model\Level\Exception\LevelNotFoundException;
use App\Context\Content\Domain\Model\Level\LevelId;
use App\Context\Content\Domain\Model\Level\LevelSlug;
use App\Context\Content\Infrastructure\Repository\Application\Query\Skill\Doctrine\SkillReadModelRepositoryDoctrine;
use App\Context\Shared\Domain\Criteria\Criteria;
use App\Context\Shared\Domain\Criteria\Field;
use App\Context\Shared\Domain\Criteria\Filter;
use App\Context\Shared\Domain\Criteria\FilterOperator;
use App\Context\Shared\Domain\Criteria\FilterStringValue;
use App\Context\Shared\Infrastructure\Doctrine\Criteria\CriteriaConvert;
use Doctrine\DBAL\Query\QueryBuilder;
use Doctrine\ORM\EntityManagerInterface;

class LevelReadModelRepositoryDoctrine implements LevelReadRepository
{
    public function __construct(
        private readonly EntityManagerInterface $em,
        private readonly SkillReadModelRepositoryDoctrine $skillReader
    ) {
    }

    public function byId(string $skillId, string $levelId): LevelReadModel
    {
        $this->skillReader->byId($skillId);

        $criteria = new Criteria();
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('skill_id'),
                    FilterOperator::EQ,
                    new FilterStringValue($skillId),
                ),
            );
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('id'),
                    FilterOperator::EQ,
                    new FilterStringValue($levelId),
                ),
            );
        $criteria->setLimit(1);

        $level = current($this->getList($criteria));
        if ($level) {
            return $level;
        }

        throw new LevelNotFoundException(new LevelId($levelId));
    }

    public function bySlug(string $skillSlug, string $levelSlug): LevelReadModel
    {
        $skill = $this->skillReader->bySlug($skillSlug);

        $criteria = new Criteria();
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('skill_id'),
                    FilterOperator::EQ,
                    new FilterStringValue($skill->id),
                ),
            );
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('slug'),
                    FilterOperator::EQ,
                    new FilterStringValue($levelSlug),
                ),
            );
        $criteria->setLimit(1);

        $level = current($this->getList($criteria));
        if ($level) {
            return $level;
        }

        throw new LevelNotFoundBySlugException(new LevelSlug($levelSlug));
    }

    /**
     * @return LevelReadModel[]
     *
     * @throws \Doctrine\DBAL\Exception
     *
     * @psalm-suppress MixedReturnTypeCoercion
     */
    public function getList(?Criteria $criteria = null): array
    {
        $qb = new QueryBuilder($this->em->getConnection());
        $qb->from('level', 'l')
            ->select('l.id, l.skill_id, l.title, l.slug, l.description, l.count_need_approvals')
            ->where('l.is_removed = false')
            ->addSelect('(SELECT count(*) FROM question q WHERE l.id = q.level_id AND q.is_removed = false) as count_questions')
            ->orderBy('l.title', 'ASC');

        if ($criteria) {
            $qb = CriteriaConvert::criteriaToQueueBuilder($qb, $criteria);
        }

        $result = $qb->executeQuery();

        return array_map($this->getFnMapping(), $result->fetchAllAssociative());
    }

    protected function getFnMapping(): \Closure
    {
        return function (array $data): LevelReadModel {
            return new LevelReadModel(
                (string) ($data['skill_id'] ?? ''),
                (string) ($data['id'] ?? ''),
                (string) ($data['title'] ?? ''),
                (string) ($data['slug'] ?? ''),
                (string) ($data['description'] ?? ''),
                (int) ($data['count_need_approvals'] ?? 0),
                (int) ($data['count_questions'] ?? 0)
            );
        };
    }
}
