<?php

declare(strict_types=1);

namespace App\Context\Content\Infrastructure\Repository\Application\Query\Question\Doctrine;

use App\Context\Content\Application\Query\Question\ReadModel\QuestionNotFoundBySlugException;
use App\Context\Content\Application\Query\Question\ReadModel\QuestionReadModel;
use App\Context\Content\Application\Query\Question\ReadModel\QuestionReadRepository;
use App\Context\Content\Domain\Model\Question\Exception\QuestionNotFoundException;
use App\Context\Content\Domain\Model\Question\QuestionId;
use App\Context\Content\Domain\Model\Question\QuestionSlug;
use App\Context\Content\Infrastructure\Repository\Application\Query\Level\Doctrine\LevelReadModelRepositoryDoctrine;
use App\Context\Shared\Domain\Criteria\Criteria;
use App\Context\Shared\Domain\Criteria\Field;
use App\Context\Shared\Domain\Criteria\Filter;
use App\Context\Shared\Domain\Criteria\FilterOperator;
use App\Context\Shared\Domain\Criteria\FilterStringValue;
use App\Context\Shared\Infrastructure\Doctrine\Criteria\CriteriaConvert;
use Doctrine\DBAL\Query\QueryBuilder;
use Doctrine\ORM\EntityManagerInterface;

class QuestionReadModelRepositoryDoctrine implements QuestionReadRepository
{
    public function __construct(
        private readonly EntityManagerInterface $em,
        private readonly LevelReadModelRepositoryDoctrine $levelReader
    ) {
    }

    public function byId(string $skillId, string $levelId, string $questionId): QuestionReadModel
    {
        $this->levelReader->byId($skillId, $levelId);

        $criteria = new Criteria();
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('level_id'),
                    FilterOperator::EQ,
                    new FilterStringValue($levelId)
                )
            );
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('id'),
                    FilterOperator::EQ,
                    new FilterStringValue($questionId)
                )
            );
        $criteria->setLimit(1);

        $question = current($this->getList($criteria));
        if ($question) {
            return $question;
        }

        throw new QuestionNotFoundException(new QuestionId($questionId));
    }

    public function bySlug(string $skillSlug, string $levelSlug, string $questionSlug): QuestionReadModel
    {
        $level = $this->levelReader->bySlug($skillSlug, $levelSlug);

        $criteria = new Criteria();
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('level_id'),
                    FilterOperator::EQ,
                    new FilterStringValue($level->id)
                )
            );
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('slug'),
                    FilterOperator::EQ,
                    new FilterStringValue($questionSlug)
                )
            );
        $criteria->setLimit(1);

        $question = current($this->getList($criteria));

        if ($question) {
            return $question;
        }

        throw new QuestionNotFoundBySlugException(new QuestionSlug($questionSlug));
    }

    /**
     * @return QuestionReadModel[]
     *
     * @throws \Doctrine\DBAL\Exception
     *
     * @psalm-suppress MixedReturnTypeCoercion
     */
    public function getList(?Criteria $criteria = null): array
    {
        $qb = new QueryBuilder($this->em->getConnection());
        $qb->from('question', 'q')
            ->select('q.id, q.level_id, q.title, q.slug, q.description, q.points')
            ->where('q.is_removed = false')
            ->orderBy('q.title', 'ASC');

        if ($criteria) {
            $qb = CriteriaConvert::criteriaToQueueBuilder($qb, $criteria);
        }

        $result = $qb->executeQuery();

        return array_map($this->getFnMapping(), $result->fetchAllAssociative());
    }

    protected function getFnMapping(): \Closure
    {
        return function (array $data): QuestionReadModel {
            return new QuestionReadModel(
                (string) ($data['level_id'] ?? ''),
                (string) ($data['id'] ?? ''),
                (string) ($data['title'] ?? ''),
                (string) ($data['slug'] ?? ''),
                (string) ($data['description'] ?? ''),
                (int) ($data['points'] ?? 0)
            );
        };
    }
}
