<?php

declare(strict_types=1);

namespace App\Context\Content\Infrastructure\Repository\Application\Query\Skill\Doctrine;

use App\Context\Content\Application\Query\Skill\ReadModel\SkillNotFoundBySlugException;
use App\Context\Content\Application\Query\Skill\ReadModel\SkillReadModel;
use App\Context\Content\Application\Query\Skill\ReadModel\SkillReadRepository;
use App\Context\Content\Domain\Model\Skill\Exception\SkillNotFoundException;
use App\Context\Content\Domain\Model\Skill\SkillId;
use App\Context\Content\Domain\Model\Skill\SkillSlug;
use App\Context\Shared\Domain\Criteria\Criteria;
use App\Context\Shared\Domain\Criteria\Field;
use App\Context\Shared\Domain\Criteria\Filter;
use App\Context\Shared\Domain\Criteria\FilterOperator;
use App\Context\Shared\Domain\Criteria\FilterStringValue;
use App\Context\Shared\Infrastructure\Doctrine\Criteria\CriteriaConvert;
use Doctrine\DBAL\Query\QueryBuilder;
use Doctrine\ORM\EntityManagerInterface;

class SkillReadModelRepositoryDoctrine implements SkillReadRepository
{
    public function __construct(
        private EntityManagerInterface $em
    ) {
    }

    public function byId(string $id): SkillReadModel
    {
        $criteria = new Criteria();
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('id'),
                    FilterOperator::EQ,
                    new FilterStringValue($id)
                )
            );
        $criteria->setLimit(1);

        $skill = current($this->getList($criteria));
        if ($skill) {
            return $skill;
        }

        throw new SkillNotFoundException(new SkillId($id));
    }

    public function bySlug(string $slug): SkillReadModel
    {
        $criteria = new Criteria();
        $criteria->filters()
            ->add(
                new Filter(
                    new Field('slug'),
                    FilterOperator::EQ,
                    new FilterStringValue($slug)
                )
            );
        $criteria->setLimit(1);

        $skill = current($this->getList($criteria));
        if ($skill) {
            return $skill;
        }

        throw new SkillNotFoundBySlugException(new SkillSlug($slug));
    }

    /**
     * @return SkillReadModel[]
     *
     * @throws \Doctrine\DBAL\Exception
     *
     * @psalm-suppress MixedReturnTypeCoercion
     */
    public function getList(?Criteria $criteria = null): array
    {
        $qb = new QueryBuilder($this->em->getConnection());
        $qb->from('skill', 's')
            ->select('s.id, s.title, s.slug, s.description')
            ->where('s.is_removed = false')
            ->addSelect('(SELECT count(*) FROM level l WHERE s.id = l.skill_id AND l.is_removed = false) as count_levels')
            ->orderBy('s.title', 'ASC');

        if ($criteria) {
            $qb = CriteriaConvert::criteriaToQueueBuilder($qb, $criteria);
        }

        $result = $qb->executeQuery();

        return array_map($this->getFnMapping(), $result->fetchAllAssociative());
    }

    protected function getFnMapping(): \Closure
    {
        return function (array $data): SkillReadModel {
            return new SkillReadModel(
                (string) ($data['id'] ?? ''),
                (string) ($data['title'] ?? ''),
                (string) ($data['slug'] ?? ''),
                (string) ($data['description'] ?? ''),
                (int) ($data['count_levels'] ?? 0)
            );
        };
    }
}
